// From David Eck and Richard Palais. http://3D-XplorMath.org/j/index.html.

//package vmm.core;

/**
 * A complex number, with a real and an imaginary part.  (Possibley to be replaced with
 * a class that has better support for complex arithmetic and functions of a complex variable.)
 */
class Complex {

    var re: Double = 0.toDouble()
    var im: Double = 0.toDouble()

    /**
     * Create a complex number initially equal to zero
     */
    constructor() {}

    /**
     * Create a complex number initially equal to the real number x.
     */
    constructor(x: Double) {
        re = x
    }

    /**
     * Create a complex number initially equal to x + iy
     */
    constructor(x: Double, y: Double) {
        re = x //.toDouble()
        im = y //.toDouble()
    }

    /**
     * Create a new complex number that is initially equal to a given complex number.
     * @param c The complex number to be copied.  If null, it is treated as zero.
     */
    constructor(c: Complex) {
        copy(c)
    }

    /**
     * Returns true if obj is equal to this complex number.  If obj is null or is not
     * of type Complex, the return value is false.
     */
    override fun equals(obj: Any?): Boolean {
        try {
            val c = obj as Complex?
            return c!!.re == re && c.im == im
        } catch (e: Exception) {
            return false
        }

    }

    /**
     * Computes the conjugate of a complex number.
     */
    fun conj(): Complex {
        return Complex(re, -im)
    }

    /**
     * Sets this complex number equal to a copy of a given number.
     * @param c The number to be copied; if null, the number is treated as zero.
     */
    fun copy(c: Complex?) {
        if (c == null) {
            im = 0.0
            re = im
        }
        else {
            re = c.re
            im = c.im
        }
    }

    /**
     * Returns this + c; c must be non-null.
     */
    operator fun plus(c: Complex): Complex {
        return Complex(re + c.re, im + c.im)
    }

    /**
     * Returns this - c; c must be non-null.
     */
    operator fun minus(c: Complex): Complex {
        return Complex(re - c.re, im - c.im)
    }

    /**
     * Returns this * c; c must be non-null.
     */
    operator fun times(c: Complex): Complex {
        return Complex(re * c.re - im * c.im, re * c.im + im * c.re)
    }

    /**
     * Returns this / c; c must be non-null.
     */
    fun dividedBy(c: Complex): Complex {
        val denom = c.re * c.re + c.im * c.im
        if (denom == 0.0)
            return Complex(java.lang.Double.NaN, java.lang.Double.NaN)
        else
            return Complex((re * c.re + im * c.im) / denom, (im * c.re - re * c.im) / denom)
    }

    operator fun times(x: Double): Complex {
        return Complex(re * x, im * x)
    }

    operator fun plus(x: Double): Complex {
        return Complex(re + x, im)
    }

    operator fun minus(x: Double): Complex {
        return Complex(re - x, im)
    }

    fun dividedBy(x: Double): Complex {
        return Complex(re / x, im / x)
    }

    /**
     * Returns the absolute value squared of this.
     * @return real part squared plus imaginary part squared
     */
    fun abs2(): Double {
        return re * re + im * im
    }

    /**
     * Returns the absolute value, "r" in polar coordinates, of this.
     * @return the square root of (real part squared plus imaginary part squared)
     */
    fun r(): Double {
        return Math.sqrt(re * re + im * im)
    }

    /**
     * Returns arg(this), the angular polar coordinate of this complex number, in the range -pi to pi.
     * The return value is simply Math.atan2(imaginary part, real part).
     */
    fun theta(): Double {
        return Math.atan2(im, re)
    }

    /**
     * Computes the complex exponential function, e^z, where z is this complex number.
     */
    fun exponential(): Complex {
        val length = Math.exp(re)
        return Complex(length * Math.cos(im), length * Math.sin(im))
    }

    /**
     * Computes the complex reciprocal function, 1/z, where z is this complex number.
     */
    fun inverse(): Complex {
        val length = re * re + im * im
        return Complex(re / length, -im / length)
    }

    fun log(): Complex {
        val modulus = Math.sqrt(re * re + im * im)
        val arg = Math.atan2(im, re)
        return Complex(Math.log(modulus), arg)
    }

    /**
     * Computes that complex logarithm of this complex number
     * that is nearest to previous.
     * A test code is in fractals.TestAnalyticContinuation.
     */
    fun logNearer(previous: Complex): Complex {
        val c = Complex(this.log())
        val h = (c.im - previous.im) / (2 * Math.PI)
        val d = 2 * Math.PI * Math.floor(h + 0.5)
        c.im = c.im - d
        return c
    }

    fun sinh(x: Double): Double {
        return (Math.exp(x) - Math.exp(-x)) / 2
    }

    fun cosh(x: Double): Double {
        return (Math.exp(x) + Math.exp(-x)) / 2
    }

    fun sine(): Complex {
        val x: Double
        val y: Double
        val z = Complex(0.0, 0.0)
        x = re
        y = im
        z.re = Math.sin(x) * cosh(y)
        z.im = Math.cos(x) * sinh(y)
        return z
    }

    fun power(x: Double): Complex {
        val modulus = Math.sqrt(re * re + im * im)
        val arg = Math.atan2(im, re)
        val log_re = Math.log(modulus)
        val log_im = arg
        val x_log_re = x * log_re
        val x_log_im = x * log_im
        val modulus_ans = Math.exp(x_log_re)
        return Complex(modulus_ans * Math.cos(x_log_im), modulus_ans * Math.sin(x_log_im))
    }

    /**
     * Returns a complex k-th root of this complex number.  The root that is returned is
     * the one with the smallest positive arg.
     * (If k is 0, the return value is 1.  If k is negative, the value is 1/integerRoot(-k).)
     */
    fun integerRoot(kr: Int): Complex {
        var k = kr
        var a: Double
        var b: Double
        var neg = false
        if (k < 0) {
            k = -k
            neg = true
        }
        if (k == 0) {
            a = 1.0
            b = 0.0
        } else if (k == 1) {
            a = re
            b = im
        } else {
            var length = r()
            var angle = theta()
            if (angle < 0)
                angle += Math.PI * 2
            length = Math.pow(length, 1.0 / k)
            angle = angle / k
            a = length * Math.cos(angle)
            b = length * Math.sin(angle)
        }
        if (neg) {
            val denom = a * a + b * b
            a = a / denom
            b = -b / denom
        }
        return Complex(a, b)
    }

    /**
     * Computes that square root of this complex number
     * that is nearer to previous than to minus previous.
     * A test code is in fractals.TestAnalyticContinuation.
     */
    fun squareRootNearer(previous: Complex): Complex {
        val c: Complex
        c = this.integerRoot(2)
        if (c.re * previous.re + c.im * previous.im < 0) {
            c.re = -c.re
            c.im = -c.im
        }
        return Complex(c.re, c.im)
    }

    fun stereographicProjection(): DoubleArray {

        val rsquare: Double
        val rsquarePlusOne: Double
        val projPoint = DoubleArray(3)
        rsquare = re * re + im * im
        rsquarePlusOne = rsquare + 1
        projPoint[0] = 2 * re / rsquarePlusOne
        projPoint[1] = 2 * im / rsquarePlusOne
        projPoint[2] = (rsquare - 1) / rsquarePlusOne
        return projPoint
    }

    companion object {

        val ZERO_C = Complex(0.0, 0.0)

        val ONE_C = Complex(1.0, 0.0)

        val I_C = Complex(0.0, 1.0)


        /**
         * Returns the complex number (r*cos(theta)) + i*(r*sin(theta)).
         */
        fun polar(r: Double, theta: Double): Complex {
            return Complex(r * Math.cos(theta), r * Math.sin(theta))
        }
    }
}